﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Selections;
using CTNet;
using PdfDocument;
using STFXmlParser;
using System.Text.RegularExpressions;

namespace CalcHelper
{
    public partial class ToolBelt : Form
    {
        //DocCalcs docCalcs;

        // Our access to the 'outside world':
        private DocFile _docFile;
        private CTClient _ctClient;
        private BoxSelections _boxSel;
        private PageArray _pa;

        public enum tbTabPage { boxSel, codeNav, fcnPick, snipPick };
        public tbTabPage tabPage;

        /*
         * ***********************
         * Specific to Box Selector:
         */

        /*
         * NOTE: enum boxSelOneShot need to follow the same order as the
         * Box Selector boxFormatComboBox's list elements. boxSelFormats
         * in ProcessSelections.cs ALSO follows this order - the two need
         * to coincide.
         */
        public enum boxSelOneShot { boxToF, boxToI, box, boxQuoted, boxToFAbs, none };
        public boxSelOneShot bsOneShotState;

        /*
         * For individually boxed posting to a variable. We store the number of
         * boxes that were used in posting to a variable. When we post from that variable
         * we check against the number of boxes we will be distributing that variable
         * to and inform the user if the lengths do not agree.
         */
        public Dictionary<string, int> indBoxedVarDict = new Dictionary<string, int>();

        /*
         * ***********************
         * Specific to Code Navigator:
         */
        public enum codeNavList { all, docFcns, comMrkrs, orphBoxes };
        public int codeNavFcnListIndex;
        public int codeNavComListIndex;
        public int codeNavOrphanListIndex;

        /*
         * ***********************
         * Specific to Snippet Picker:
         */
        //int value = _dict.Values.ElementAt(5);//ElementAt value should be <= _dict.Count - 1
        //string key = _dict.Keys.ElementAt(5);//ElementAt value should be  < =_dict.Count - 1

        //public int codeSnipCatListIndex;
        //public int codeSnipPrefixListIndex;

        /*
         * ***********************
         */

        public ToolBelt(ref BoxSelections boxSel, ref CTClient ctClient, ref DocFile docFile, ref PageArray pa, tbTabPage tp)
        {
            InitializeComponent();

            _boxSel = boxSel;
            _ctClient = ctClient;
            _docFile = docFile;
            _pa = pa;

            SetupTabPages();
            SelectTabPage(tp);

            // Subscribing to the Form's Box Selection event.
            Form1.BoxSel += Form1_CalcToolBoxSel;

            /*
             * Subscribing to Selections (ProcessSelections.cs)
             * unselection event.
             */
            BoxSelections.BoxUnSelEvent += BoxSelections_UnSel;
        }

        // More ToolBelt event code is near the bottom of this file.
        // Create the ToolBelt event itself:
        public static event EventHandler<ToolBeltArgs> ToolBeltEvent;

        // Trigger the ToolBelt event:
        protected virtual void OnToolBeltEvent(ToolBeltArgs e)
        {
            if (ToolBeltEvent != null)
            {
                ToolBeltEvent(this, e);
            }
        }

        /*
         * Called from:
         * 
         * SetupTabPages()
         * DocFile.cs: SetCodeNavLists()
         * 
         * mode - 'true' means to set the Code Navigator's dynamic list's
         *          selection mode to 'One'. 
         *  
         *        'false' means to set the selection mode to 'None'. This
         *          will prevent a list item from being automatically
         *          selected, namely, the first item.
         *  
         * These are the lists that can be dynamically added to or deleted
         * from as a result of a Refresh operation. Only the user makes a
         * selection, we do not want a selection made 'for them'. When a
         * selection is made we will 'goto' that code location in the
         * editor.
         */
        public void CodeNavListSelectMode(codeNavList listId, bool mode)
        {
            SelectionMode selMode = (mode ? SelectionMode.One : SelectionMode.None);

            switch (listId)
            {
                case (codeNavList.docFcns):

                    docFcnList.SelectionMode = selMode;

                    break;

                case (codeNavList.comMrkrs):

                    comMarkerList.SelectionMode = selMode;

                    break;

                case (codeNavList.orphBoxes):

                    orphanBoxList.SelectionMode = selMode;

                    break;

                default:

                    docFcnList.SelectionMode = selMode;
                    comMarkerList.SelectionMode = selMode;
                    orphanBoxList.SelectionMode = selMode;

                    break;
            }
        }

        /*
         * Called from:
         * 
         * Pdf.cs - SetCodeNavLists().
         * 
         * We try to reset the selections to what they were before
         * the refresh. We say 'try' because the lists may change
         * in length, in which case we reset the selections
         * accordingly.
         */
        public void SetCodeNavListSel(codeNavList listId)
        {
            switch (listId)
            {
                case (codeNavList.docFcns):
                    
                    if (codeNavFcnListIndex > docFcnList.Items.Count - 1)
                    {
                        codeNavFcnListIndex = docFcnList.Items.Count - 1;
                    }

                    docFcnList.SelectedIndex = codeNavFcnListIndex;

                    break;

                case (codeNavList.comMrkrs):

                    if (codeNavComListIndex > comMarkerList.Items.Count - 1)
                    {
                        codeNavComListIndex = comMarkerList.Items.Count - 1;
                    }

                    comMarkerList.SelectedIndex = codeNavComListIndex;

                    break;

                case (codeNavList.orphBoxes):

                    if (codeNavOrphanListIndex > orphanBoxList.Items.Count - 1)
                    {
                        codeNavOrphanListIndex = orphanBoxList.Items.Count - 1;
                    }

                    orphanBoxList.SelectedIndex = codeNavOrphanListIndex;

                    break;
            }
        }

        /*
         * Called From:
         * 
         * ToolBelt() - Constructor.
         * SendOffLeftBoxSel()
         * boxSelToolStripMenuItem_Click()
         * codeNavToolStripMenuItem_Click()
         * fcnPickToolStripMenuItem_Click()
         * snipPickToolStripMenuItem_Click()
         */
        public void SelectTabPage(tbTabPage tp)
        {
            tabPage = tp;

            switch (tp)
            {
                case (tbTabPage.boxSel):

                    toolBeltTabCtrl.SelectedTab = boxSelTab;

                    break;

                case (tbTabPage.codeNav):

                    toolBeltTabCtrl.SelectedTab = codeNavTab;

                    break;

                case (tbTabPage.fcnPick):

                    toolBeltTabCtrl.SelectedTab = fcnPickTab;

                    break;

                case (tbTabPage.snipPick):

                    toolBeltTabCtrl.SelectedTab = snipPickTab;

                    break;
            }
        }

        /*
         * Our handler for Selections [ProcessSelections.cs] box
         * unselection event.
         * 
         * ResetSelections() fires this event.
        */
        public void BoxSelections_UnSel(object sender, BoxUnSelArgs e)
        {
            List<string> opWaitList = new List<string>();

            if (e != null)
            {
                if (e.boxes != null)
                {
                    if (e.boxes.Count > 0)
                    {
                        /*
                         * Go to the Box Selector tab page if not the
                         * current tab page since these unselections
                         * are boxes that the Box Selector is waiting on.
                         */
                        if (tabPage != tbTabPage.boxSel)
                        {
                            SelectTabPage(tbTabPage.boxSel);
                        }

                        /*
                         * Remove these boxes from the Box Selector's
                         * text box - we WERE waiting for the user to
                         * select an operator/separator to apply to
                         * these boxes.
                         */
                        opWaitList = boxSelTextBox.Text.Replace(" ", "").Split(',').ToList();

                        foreach (string box in e.boxes)
                        {
                            /*
                             * Should be in here, but to be sure:
                             */
                            if (opWaitList.Contains(box))
                            {
                                opWaitList.Remove(box);
                            }
                        }

                        /*
                         * Keep a tight fit with the boxSelTextBox, i.e. 1:1.
                         */
                        _boxSel.toolBeltOpWaitBoxes.Clear();

                        // Restore the textbox with what is left:
                        if (opWaitList.Count > 0)
                        {
                            boxSelTextBox.Text = string.Join(", ", opWaitList);

                            // Safe because we cleared the list above.
                            _boxSel.toolBeltOpWaitBoxes.AddRange(opWaitList);
                        }
                        else
                        {
                            /*
                             * Nothing left to wait on. 'false' because we cleared
                             * the _boxSel.toolBeltOpWaitBoxes list above.
                             */
                            ResetBoxSelOpWait(false);
                        }
                    }
                }
            }
        }

        /*
         * Our handler for the forms's box selection event.
         * 
         * Called From:
         * 
         * SendOffLeftBoxSel() - Methods.cs
        */
        public void Form1_CalcToolBoxSel(object sender, Form1.BoxSelArgs e)
        {
            string msg = "";

            if (e != null)
            {
                if (e.boxes != null)
                {
                    if (e.boxes.Count > 0)
                    {
                        /*
                         * We should already be on the Box Selector tab page but...
                         */
                        if (tabPage != tbTabPage.boxSel)
                        {
                            SelectTabPage(tbTabPage.boxSel);
                        }

                        /*
                         * The caller has done all the prep work for us
                         * and has pretty much laid out what to do here.
                         */
                        switch (e.cmd)
                        {
                            case ("MultiBoxWait"):

                                /*
                                 * A multi box selection, wait for operator/separator
                                 * selection from user - we may already be waiting from
                                 * prior multi box selection(s). No operator lock.
                                 */
                                ProcessBoxSelWait(e.boxes);

                                break;

                            case ("MultiBoxOpLockOneShot"):

                                /*
                                 * A multi box selection, operator lock in place
                                 * and we have a one shot button set. Should do 
                                 * these idiot checks though this should never
                                 * happen. We check because we need to select the
                                 * proper operator and box format.
                                 */
                                if (_boxSel.bsOpLockState == boxSelOpLock.none)
                                {
                                    msg = "Multi Box - Operator Lock State Error:\n\nTool Belt Box Selector received box selections expecting the Operator Lock to be set, however it is not set.\n\nBoxes will be unselected and not processed.";
                                }
                                else if (bsOneShotState == boxSelOneShot.none)
                                {
                                    msg = "Multi Box - Format One Shot Error:\n\nTool Belt Box Selector received box selections expecting a Box Format One Shot button to be set, however that is not the case.\n\nBoxes will be unselected and not processed.";
                                }

                                if (msg.Length > 0)
                                {
                                    ToolBeltArgs toolBeltArgs = new ToolBeltArgs();

                                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                                    {
                                        mainMsg.ShowDialog(this);
                                    }

                                    toolBeltArgs.eventType = tbEvent.clearBoxSel;
                                    toolBeltArgs.boxSel = new List<string>(e.boxes);

                                    OnToolBeltEvent(toolBeltArgs);
                                }
                                else
                                {
                                    ProcessMultiBoxOpLockOneShot(e.boxes);
                                }

                                break;

                            case ("SingleBoxWait"):

                                /*
                                 * A single box selection, add to the operator
                                 * waiting list. One shot does not matter, could
                                 * be on could be off.
                                 */
                                ProcessBoxSelWait(e.boxes);

                                break;

                            case ("SingleBoxOneShot"):

                                /*
                                 * A single box selection, have a one shot button
                                 * ready to use on this box. Not waiting on an
                                 * operator. Need to error check the one shot state
                                 * so we can select the proper box format.
                                 */
                                if (bsOneShotState == boxSelOneShot.none)
                                {
                                    ToolBeltArgs toolBeltArgs = new ToolBeltArgs();

                                    msg = "Single Box - Box Format One Shot Error:\n\nTool Belt Box Selector received a box selection expecting a Box Format One Shot button to be set, however that is not the case.\n\nBox will be unselected and not processed.";
                                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                                    {
                                        mainMsg.ShowDialog(this);
                                    }

                                    toolBeltArgs.eventType = tbEvent.clearBoxSel;
                                    toolBeltArgs.boxSel = new List<string>(e.boxes);

                                    OnToolBeltEvent(toolBeltArgs);
                                }
                                else
                                {
                                    ProcessSingleBoxOneShot(e.boxes[0]);
                                }

                                break;
                        }
                    }
                }
            }
        }

        /*
         * What has happened:
         * 
         * 1) User left mouse selected a single box;
         * 2) Tool Belt is 'up' but not waiting on an operator button;
         * 3) Tool Belt's Box Selector has a one shot box format button 'on'.
         * 
         * We would not be here if 3) above was not true, even if 2) was.
         * 
         * We will process the box using the locked operator and will
         * format it using the one shot button. We will send the
         * resultant string on to Vscode and reset the one shot button
         * to 'off'.
         * 
         * Called From:
         * 
         *      Form1_CalcToolBoxSel()
         */
        public void ProcessSingleBoxOneShot(string box)
        {
            string formatStr = "";
            string boxStr = "";
            int formatIndex = 0;

            /*
             * Get the formatStr from the format hold button
             * via bsOneShotState.
             */
            formatIndex = Array.IndexOf(Enum.GetValues(bsOneShotState.GetType()), bsOneShotState);

            formatStr = _boxSel.boxSelFormats[formatIndex];

            boxStr = string.Format(formatStr, box);

            // Turn 'off' the format hold button.
            ResetOneShotButton();

            _ctClient.SendBoxSelections(boxStr);
        }

        /*
         * What has happened:
         * 
         * 1) User left mouse selected more than one box;
         * 2) Tool Belt's Box Selector operator lock is on;
         * 3) Tool Belt is 'up';
         * 4) Tool Belt's Box Selector has a one shot box format button 'on'.
         * 
         * We would not be here if 4) above was not true, even if 3) was.
         * 
         * The way the operator lock feature works means there would be
         * no current boxes waiting in the text box hold area as we would
         * automatically process any multi box selections immediately
         * using the operator that is locked 'on'. The form sent us the
         * boxes to process.
         * 
         * We will process the boxes using the locked operator and will
         * format them using the one shot button. We will send the
         * resultant string on to Vscode and reset the one shot button
         * to 'off'.
         * 
         * Called From:
         * 
         *      Form1_CalcToolBoxSel()
         */
        public void ProcessMultiBoxOpLockOneShot(List<string> boxes)
        {
            string formatStr = "";
            string opStr = "";
            string boxStr = "";
            int formatIndex = 0;

            /*
             * We will compose the multi box selection string using
             * the locked operator and the Tool Belt's Box
             * Selector's current one shot box format and send it on
             * to Vscode via ctClient.
             * 
             * First set the operator string:
             */
            switch (_boxSel.bsOpLockState)
            {
                case (boxSelOpLock.add):
                    opStr = " + ";
                    break;

                case (boxSelOpLock.div):
                    opStr = " / ";
                    break;

                case (boxSelOpLock.mult):
                    opStr = " * ";
                    break;

                case (boxSelOpLock.sub):
                    opStr = " - ";
                    break;
            }

            /*
             * Get the formatStr from the box format hold button
             * via bsOneShotState.
             */
            formatIndex = Array.IndexOf(Enum.GetValues(bsOneShotState.GetType()), bsOneShotState);

            formatStr = _boxSel.boxSelFormats[formatIndex];

            for (int i = 0; i < boxes.Count; i++)
            {
                if (i < boxes.Count - 1)
                {
                    boxStr += (string.Format(formatStr, boxes[i]) + opStr);
                }
                else
                {
                    boxStr += string.Format(formatStr, boxes[i]);
                }
            }

            // Turn 'off' the format hold button.
            ResetOneShotButton();

            _ctClient.SendBoxSelections(boxStr);
        }

        /*
         * This is a 'Box Selector' thing.
         * 
         * What has happened:
         * 
         * [
         * Either: 1) User clicked an Operator Lock button and there
         *              was NO prior Operator Lock in place;
         * OR: 2) User clicked an Operator/Separator button;
         * ]
         * AND: There are currently boxes being held waiting for 1)
         * or 2) to happen.
         * 
         * We will process the boxes being held using the operator
         * argument and will format them as follows:
         * 
         * If a one shot box format button is set we will use that
         * box format else we will use the current box format.
         * 
         * We will send the resultant string on to Vscode and clear the
         * operator wait text box and associated list. And if a one shot
         * box format situation, we will reset that to 'off'.
         * 
         * Called From:
         * 
         *      ProcessOpLockButton()
         *      boxSelPlusBtn_Click()
         *      boxSelMinusBtn_Click()
         *      boxSelMultBtn_Click()
         *      boxSelDivBtn_Click()
         *      boxSelCommaBtn_Click()
         */
        public void ProcessOpWaitBoxes(string opStr)
        {
            string formatStr = "";
            string boxStr = "";
            int formatIndex;

            /*
             * Get the index to use for the formatStr:
             */
            if (bsOneShotState == boxSelOneShot.none)
            {
                formatIndex = _boxSel.boxSelFormatsIndex;
            }
            else
            {
                formatIndex = Array.IndexOf(Enum.GetValues(bsOneShotState.GetType()), bsOneShotState);

                // Turn 'off' the format hold button.
                ResetOneShotButton();
            }

            formatStr = _boxSel.boxSelFormats[formatIndex];

            for (int i = 0; i < _boxSel.toolBeltOpWaitBoxes.Count; i++)
            {
                if (i < _boxSel.toolBeltOpWaitBoxes.Count - 1)
                {
                    boxStr += (string.Format(formatStr, _boxSel.toolBeltOpWaitBoxes[i]) + opStr);
                }
                else
                {
                    boxStr += string.Format(formatStr, _boxSel.toolBeltOpWaitBoxes[i]);
                }
            }

            /*
             * Reset the Box Selector's operator wait state.
             */
            ResetBoxSelOpWait(true);

            _ctClient.SendBoxSelections(boxStr);
        }

        /*
         * This places the list of boxes into the Box Selector's text box
         * holding area and the associated list.
         * 
         * The Box Selector is/will be waiting for the user to select an
         * operator or separator button to apply to these boxes before
         * sending the resultant string to the Vscode editor file.
         * 
         * It has already been verified that there is no operator lock
         * in effect.
         * 
         * Caller has ensured there is at least one box in boxes.
         * 
         * If this is the first multi box selection we will prompt the
         * user via message text in the label area above the text box.
         * 
         * Called From:
         * 
         *      Form1_CalcToolBoxSel()
         */
        public void ProcessBoxSelWait(List<string> boxes)
        {
            if (_boxSel.toolBeltOpWaitBoxes.Count == 0)
            {
                boxSelMsg1Label.Text = "Option1: Click an 'Operator Lock' or an 'Operators/Separators' Button to apply to these boxes";
                boxSelMsg2Label.Text = "Option2: Use the 'Individually Boxed Calcs' area if these are individually boxed boxes";
                boxSelMsg3Label.Text = "Click the 'Abort Selections' button to cancel these selections";

                boxSelAbortBtn.Enabled = true;
                indBoxedGrpBox.Enabled = true;
            }

            boxSelTextBox.AppendText((boxSelTextBox.Text.Length > 0 ? ", " : "") + string.Join(", ", boxes));
            _boxSel.toolBeltOpWaitBoxes.AddRange(boxes);

            /* 
             * Want to catch key presses if they use the keyboard for
             * operator/separator selection.
             */
            this.Focus();
        }

        /*
         * This 'resets' the Box Selector's operator wait
         * state.
         * 
         * This is the state where multiple box selections
         * were originally received by the Box Selector and there
         * was no operator lock in effect, so we were holding on
         * to these boxes waiting for the user to select either an
         * operator lock button or a operator/separator button
         * at which time we would apply this operator/separator
         * to the held boxes and then send them on to Vscode for
         * insertion at the current cursor location.
         * 
         * We may have also just processed boxes for an individually
         * boxed calculation.
         * 
         * Called From:
         * 
         *      BoxSelections_UnSel()
         *      boxSelAbortBtn_Click()
         *      boxSelClearAllBtn_Click()
         *      ProcessOpWaitBoxes()
         *      indBoxCalcToVarBtn_Click()
         *      
         * Arguments:
         * 
         *      clearWaitList - If true we will clear the
         *                  _boxSel.toolBeltOpWaitBoxes.
         */
        public void ResetBoxSelOpWait(bool clearWaitList)
        {
            boxSelMsg1Label.Text = "";
            boxSelMsg2Label.Text = "";
            boxSelMsg3Label.Text = "";

            boxSelTextBox.Clear();

            if (clearWaitList == true)
            {
                _boxSel.toolBeltOpWaitBoxes.Clear();
            }

            boxSelAbortBtn.Enabled = false;
            indBoxedGrpBox.Enabled = false;
        }

        /*
         * This 'turns off' the Box Selector's one shot
         * box format button.
         * 
         * Called From:
         * 
         *      ProcessMultiBoxOpLockOneShot()
         */
        private void ResetOneShotButton()
        {
            switch (bsOneShotState)
            {
                case (boxSelOneShot.box):
                    boxNoFormatToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    break;

                case (boxSelOneShot.boxQuoted):
                    boxUnformatQuotedToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    break;

                case (boxSelOneShot.boxToF):
                    box_to_fToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    break;

                case (boxSelOneShot.boxToFAbs):
                    box_to_f_absToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    break;

                case (boxSelOneShot.boxToI):
                    box_to_iToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    break;
            }

            bsOneShotState = boxSelOneShot.none;
        }

        /*
         * This processess the 'Operator Lock - Multi Box' button clicks
         * on the 'Box Selector' tab page. We make sure only one operator
         * lock button can be selected at a time.
         * 
         * If this is a locking operation where there is NO current
         * operator button locked:
         * 
         *      We will check if there are boxes being held as we were waiting
         *      for the user to select an operator/separator button, if so
         *      we will call a function that will use the operator being locked
         *      in building the box string to be sent on to Vscode. The called
         *      function will then empty the box operator hold text box and
         *      associated list.
         *      
         * Any future multi box left mouse button selections will use this
         * operator in building the box string for sending on to Vscode -
         * this will happen even if the Tool Box is 'down' as this setting
         * persists. What this means is there will be no need to move the
         * boxes into the operator hold area because we have a locked
         * operator for multi box selections.
         * 
         * Called From:
         * 
         *      boxSelOpLockDivBtn_Click()
         *      boxSelOpLockMultBtn_Click()
         *      boxSelOpLockMinusBtn_Click()
         *      boxSelOpLockPlusBtn_Click()
         */
        private void ProcessOpLockButton(boxSelOpLock opLockBtn)
        {
            bool opLockOn = true;
            bool wasNoLock = (_boxSel.bsOpLockState == boxSelOpLock.none);

            if (wasNoLock == true)
            {
                _boxSel.bsOpLockState = opLockBtn;
            }
            else if (_boxSel.bsOpLockState == opLockBtn)
            {
                // Resetting the current operator lock button:
                _boxSel.bsOpLockState = boxSelOpLock.none;
                opLockOn = false;
            }
            else
            {
                // 'Turn off' the present operator lock button:
                switch (_boxSel.bsOpLockState)
                {
                    case (boxSelOpLock.div):
                        boxSelOpLockDivBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;

                    case (boxSelOpLock.mult):
                        boxSelOpLockMultBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;

                    case (boxSelOpLock.sub):
                        boxSelOpLockMinusBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;

                    case (boxSelOpLock.add):
                        boxSelOpLockPlusBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;
                }

                // Set to the new operator lock button:
                _boxSel.bsOpLockState = opLockBtn;
            }

            switch (opLockBtn)
            {
                case (boxSelOpLock.div):

                    if (opLockOn == true)
                    {
                        boxSelOpLockDivBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        boxSelOpLockDivBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;

                case (boxSelOpLock.mult):

                    if (opLockOn == true)
                    {
                        boxSelOpLockMultBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        boxSelOpLockMultBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;

                case (boxSelOpLock.sub):

                    if (opLockOn == true)
                    {
                        boxSelOpLockMinusBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        boxSelOpLockMinusBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;

                case (boxSelOpLock.add):

                    if (opLockOn == true)
                    {
                        boxSelOpLockPlusBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        boxSelOpLockPlusBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;
            }

            if ((wasNoLock == true) && (_boxSel.toolBeltOpWaitBoxes.Count > 0))
            {
                string opStr = "";

                /*
                 * Technically possible the count is 1 due to user unselecting
                 * boxes that were waiting for an operator/separator selection.
                 */
                if (_boxSel.toolBeltOpWaitBoxes.Count > 1)
                {
                    switch (opLockBtn)
                    {
                        case (boxSelOpLock.add):
                            opStr = " + ";
                            break;

                        case (boxSelOpLock.div):
                            opStr = " / ";
                            break;

                        case (boxSelOpLock.mult):
                            opStr = " * ";
                            break;

                        case (boxSelOpLock.sub):
                            opStr = " - ";
                            break;
                    }
                }

                ProcessOpWaitBoxes(opStr);
            }
        }

        /*
         * This sets the Box Selector's Operator Lock button when the Tool
         * Belt starts up.
         * 
         * Called From:
         * 
         *      SetupTabPages()
         */
        private void SetupBoxSelOpLock()
        {
            switch (_boxSel.bsOpLockState)
            {
                case (boxSelOpLock.div):

                    boxSelOpLockDivBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    break;

                case (boxSelOpLock.mult):

                    boxSelOpLockMultBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    break;

                case (boxSelOpLock.sub):

                    boxSelOpLockMinusBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    break;

                case (boxSelOpLock.add):

                    boxSelOpLockPlusBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    break;
            }
        }

        /*
         * This processess the one-shot box format button clicks on the
         * 'Box Selector' tab page. We make sure only one button can be
         * selected at a time.
         * 
         * This one-shot indication is automatically reset after it is used 
         * in sending box selections to the Vscode file. What we are doing
         * here is managing multiple clicks before the one-shot selection
         * is used.
         * 
         * Called From:
         * 
         *      box_to_fToolBtn_Click()
         *      box_to_iToolBtn_Click()
         *      boxNoFormatToolBtn_Click()
         *      boxUnformatQuotedToolBtn_Click()
         *      box_to_f_absToolBtn_Click()
         */
        private void ProcessOneShotButton(boxSelOneShot oneShotBtn)
        {
            bool oneShotOn = true;

            if (bsOneShotState == boxSelOneShot.none)
            {
                bsOneShotState = oneShotBtn;
            }
            else if (bsOneShotState == oneShotBtn)
            {
                // Resetting the current one shot button:
                bsOneShotState = boxSelOneShot.none;
                oneShotOn = false;
            }
            else
            {
                // 'Turn off' the present one shot button:
                switch (bsOneShotState)
                {
                    case (boxSelOneShot.box):
                        boxNoFormatToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;

                    case (boxSelOneShot.boxQuoted):
                        boxUnformatQuotedToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;

                    case (boxSelOneShot.boxToF):
                        box_to_fToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;

                    case (boxSelOneShot.boxToFAbs):
                        box_to_f_absToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;

                    case (boxSelOneShot.boxToI):
                        box_to_iToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                        break;
                }

                // Set to the new one-shot button:
                bsOneShotState = oneShotBtn;
            }

            switch (oneShotBtn)
            {
                case (boxSelOneShot.box):

                    if (oneShotOn == true)
                    {
                        boxNoFormatToolBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        boxNoFormatToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;

                case (boxSelOneShot.boxQuoted):

                    if (oneShotOn == true)
                    {
                        boxUnformatQuotedToolBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        boxUnformatQuotedToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;

                case (boxSelOneShot.boxToF):

                    if (oneShotOn == true)
                    {
                        box_to_fToolBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        box_to_fToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;

                case (boxSelOneShot.boxToFAbs):

                    if (oneShotOn == true)
                    {
                        box_to_f_absToolBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        box_to_f_absToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;

                case (boxSelOneShot.boxToI):

                    if (oneShotOn == true)
                    {
                        box_to_iToolBtn.BackColor = Color.FromKnownColor(KnownColor.Lime);
                    }
                    else
                    {
                        box_to_iToolBtn.BackColor = Color.FromKnownColor(KnownColor.ControlLight);
                    }

                    break;
            }
        }

        /*
         * This sets up all of the ToolBelt's tab pages upon tool
         * instantiation.
         * 
         * Called From:
         * 
         *      ToolBelt() constructor.
         */
        private void SetupTabPages()
        {
            // -----------------------------------------
            // Box Selector tab page setup:
            if (_docFile.newBoxFormat == false)
            {
                // Set the boxFormatComboBox list to the old box format:
                boxFormatComboBox.Items.Clear();

                // IMPORTANT: Keep this order. Refer to comment above 
                // for public enum boxSelOneShot.
                boxFormatComboBox.Items.Add("box1_1.to_f()");
                boxFormatComboBox.Items.Add("box1_1.to_i()");
                boxFormatComboBox.Items.Add("box1_1");
                boxFormatComboBox.Items.Add("\"box1_1\"");
                boxFormatComboBox.Items.Add("box1_1.to_f().abs()");

                // Set the old format text for the one shot buttons:
                box_to_fToolBtn.Text = "box1_1.to_f()";
                box_to_iToolBtn.Text = "box1_1.to_i()";
                boxNoFormatToolBtn.Text = "box1_1";
                boxUnformatQuotedToolBtn.Text = "\"box1_1\"";
                box_to_f_absToolBtn.Text = "box1_1.to_f().abs()";
            }

            boxFormatComboBox.SelectedIndex = _boxSel.boxSelFormatsIndex;
            indBoxPrecComboBox.SelectedIndex = 2;
            bsOneShotState = boxSelOneShot.none;

            if (_boxSel.bsOpLockState != boxSelOpLock.none)
            {
                SetupBoxSelOpLock();
            }

            // -----------------------------------------
            // Code Navigator tab page setup:

            foreach (string item in _docFile.ctrlArrays)
            {
                if (item.Contains("getCheckBoxList"))
                {
                    ctrlArrayList.Items.Add("Check Box Control");
                }
                else if (item.Contains("calc_controlled"))
                {
                    ctrlArrayList.Items.Add("Calc Controlled");
                }
            }

            // Turn off the Code Navigator's dynamic lists selections.
            CodeNavListSelectMode(codeNavList.all, false);

            // The Datasources are modified by SetCodeNavLists() in Pdf.cs:
            docFcnList.DataSource = _docFile.codeNavDocFcns;
            comMarkerList.DataSource = _docFile.codeNavComMarkers;
            orphanBoxList.DataSource = _docFile.codeNavOrphanBoxes;

            // Turn on the Code Navigator's dynamic lists selections.
            CodeNavListSelectMode(codeNavList.all, true);

            codeNavFcnListIndex = -1;
            codeNavComListIndex = -1;
            codeNavOrphanListIndex = -1;

            // -----------------------------------------
            // Function Picker tab page setup:
            if (_docFile.fcnSnipPickDict.Count > 0)
            {
                foreach (KeyValuePair<string, FunctionPicker> fcn in _docFile.fcnSnipPickDict)
                {
                    fcnNameList.Items.Add(fcn.Key);
                }

                fcnNameList.SetSelected(0, true);
            }
            else
            {
                fcnSnipPickSelLbl.Text = "Calc Helper was not able to load the Function Picker code snippets";
            }

            // -----------------------------------------
            // Snippet Picker tab page setup:
            if (_docFile.snipPickDict.Count > 0)
            {
                foreach (KeyValuePair<string, SnippetPicker> snipCat in _docFile.snipPickDict)
                {
                    snipCatList.Items.Add(snipCat.Value.catName);
                }

                snipCatList.SetSelected(0, true);
            }
            else
            {
                snipPickSelLbl.Text = "Calc Helper was not able to load the Snippet Picker code snippets";
            }
        }

        private void ToolBelt_Load(object sender, EventArgs e)
        {
            if (Owner != null)
            {
                // Center CalcTool window in Parent form.
                Location = new Point(Owner.Location.X + Owner.Width / 2 - Width / 2,
                    Owner.Location.Y + Owner.Height / 2 - Height / 2);
            }
        }

        private void toolBeltExitBtn_Click(object sender, EventArgs e)
        {
            /*
             * We un-subscribe from the Form1.CalcToolBoxSel and
             * BoxSelections_UnSel events in the Tool Belt
             * FormClosed event handler 'ToolBeltClosed' located
             * in Form1.cs.
            */
            this.Close();
        }

        private void toolBeltTabCtrl_Selected(object sender, TabControlEventArgs e)
        {
            // First determine the selected tab page by using the index value.
            switch (toolBeltTabCtrl.SelectedIndex)
            {
                case (0):

                    // Box Selector tab clicked:
                    //SetupTabPages(tbTabPage.boxSel);
                    tabPage = tbTabPage.boxSel;

                    break;

                case (1):

                    // Code Navigator tab clicked:
                    //SetupTabPages(tbTabPage.codeNav);
                    tabPage = tbTabPage.codeNav;

                    break;

                case (2):

                    // Function Picker tab clicked:
                    //SetupTabPages(tbTabPage.fcnPick);
                    tabPage = tbTabPage.fcnPick;

                    break;

                case (3):

                    // Snippet Picker tab clicked:
                    //SetupTabPages(tbTabPage.snipPick);
                    tabPage = tbTabPage.snipPick;

                    break;
            }
        }

        private void ctrlArrayList_Click(object sender, EventArgs e)
        {
            if (ctrlArrayList.SelectedIndex != -1)
            {
                /*
                 * Can be found regardless of find widget using regular expression
                 * mode or not using it. We close the find widget as there is only
                 * one of these critters in the code file.
                 */
                _ctClient.GotoCode(true, false, _docFile.ctrlArrays[ctrlArrayList.SelectedIndex]);
            }
        }

        private void docFcnList_Click(object sender, EventArgs e)
        {
            codeNavFcnListIndex = docFcnList.SelectedIndex;

            if (docFcnList.SelectedIndex != -1)
            {
                //MessageBox.Show(new Form { TopMost = true }, "Click/Selection!", "docFcnList_Click():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                /*
                 * We know we have a left paren here else we would not have any
                 * function declarations in this list. See DocFile.cs -
                 * ProcessJSFile() for the regex used. Now we can search regardless
                 * of the find widget using or not using regular expression mode,
                 * and we close the find as there is only one function declaration.
                 */
                string fcnStr = _docFile.docFunctions[docFcnList.SelectedIndex];

                _ctClient.GotoCode(true, false, fcnStr.Substring(0, fcnStr.IndexOf('(')));
            }
        }

        private void comMarkerList_Click(object sender, EventArgs e)
        {
            codeNavComListIndex = comMarkerList.SelectedIndex;

            if (comMarkerList.SelectedIndex != -1)
            {
                //MessageBox.Show(new Form { TopMost = true }, "Click/Selection!", "comMarkerList_Click():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                /*
                 * These have been found in code file regardless of the find widget using
                 * the regular expression mode or not. So closing the find widget as there
                 * is only one location in the code file for each of these.
                 */
                _ctClient.GotoCode(true, false, _docFile.comMarkers[comMarkerList.SelectedIndex]);
            }
        }

        private void orphanBoxList_Click(object sender, EventArgs e)
        {
            codeNavOrphanListIndex = orphanBoxList.SelectedIndex;

            if (orphanBoxList.SelectedIndex != -1)
            {
                //MessageBox.Show(new Form { TopMost = true }, "Click/Selection!", "orphanBoxList_Click():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                /*
                 * We want the find widget to stay open in case an orphan box is in
                 * more than one location in the code file.
                 */
                _ctClient.GotoCode(false, false, _docFile.orphanBoxes[orphanBoxList.SelectedIndex]);
            }
        }

        /*
         * One Shot box format buttons - START:
         * 
         * We are either going to activate ONE of these boxes for the
         * one-shot format it is tied to, so the foreground color will
         * change to indicate it is 'armed' or reset the box to the
         * dormant state where it's foreground color is the normal color.
         * 
         * If set for the one-shot operation: Any box selections will
         * use the box's format when being sent on to Vscode, after which
         * we will reset the one-shot state.
         *
         * This is the click event handler for the '.to_f()' One-Shot
         * format button.
         */
        private void box_to_fToolBtn_Click(object sender, EventArgs e)
        {
            ProcessOneShotButton(boxSelOneShot.boxToF);
        }

        /*
         * This is the click event handler for the '.to_i()' One-Shot
         * format button.
         */
        private void box_to_iToolBtn_Click(object sender, EventArgs e)
        {
            ProcessOneShotButton(boxSelOneShot.boxToI);
        }

        /*
         * This is the click event handler for the 'no format' One-Shot
         * format button, so box1 stays at box1.
         */
        private void boxNoFormatToolBtn_Click(object sender, EventArgs e)
        {
            ProcessOneShotButton(boxSelOneShot.box);
        }

        /*
         * This is the click event handler for the 'quoted no format'
         * One-Shot format button, so box1 becomes "box1".
         */
        private void boxUnformatQuotedToolBtn_Click(object sender, EventArgs e)
        {
            ProcessOneShotButton(boxSelOneShot.boxQuoted);
        }

        /*
         * This is the click event handler for the '.to_f().abs()'
         * One-Shot format button.
         */
        private void box_to_f_absToolBtn_Click(object sender, EventArgs e)
        {
            ProcessOneShotButton(boxSelOneShot.boxToFAbs);
        }
        /*
         * One Shot box format buttons - STOP.
         */

        /*
         * This will clear ALL box selections on the document
         * that are being displayed in the Box Selector's 
         * operator wait text box.
         * 
         * We were waiting for the user to indicate what
         * operator/separator to use with these boxes, but
         * they have aborted the selections.
         * 
         * We fire an event off to the form for this.
         * 
         * We will reset the operator wait state via
         * ResetBoxSelOpWait() but let ResetSelections()
         * in ProcessSelections.cs clear the 
         * _boxSel.toolBeltOpWaitBoxes list after it is
         * finished deleting the boxes from the
         * leftBoxSelDict dictionary.
         */
        private void boxSelAbortBtn_Click(object sender, EventArgs e)
        {
            ToolBeltArgs toolBeltArgs = new ToolBeltArgs();

            ResetBoxSelOpWait(false);

            toolBeltArgs.eventType = tbEvent.clearOpWaitSel;
            OnToolBeltEvent(toolBeltArgs);
        }

        /*
         * This will clear ALL box selections on the document.
         * We fire an event off to the form for this.
         * 
         * We will reset the operator wait state via
         * ResetBoxSelOpWait() which will also clear the 
         * associated _boxSel.toolBeltOpWaitBoxes.
         * 
         * We can empty the list here because it will not
         * be needed when ResetSelections() unselects ALL
         * selections in the leftBoxSelDict.
         */
        private void boxSelClearAllBtn_Click(object sender, EventArgs e)
        {
            ToolBeltArgs toolBeltArgs = new ToolBeltArgs();

            ResetBoxSelOpWait(true);

            toolBeltArgs.eventType = tbEvent.clearAllSel;
            OnToolBeltEvent(toolBeltArgs);
        }

        private void boxFormatComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            _boxSel.boxSelFormatsIndex = boxFormatComboBox.SelectedIndex;
        }

        /*
         * We send the [ key to Vscode and it automatically adds
         * [|] to the current cursor location where | represents
         * the cursor..just as if you manually pressed the [ key.
         */
        private void grpBracketBtn_Click(object sender, EventArgs e)
        {
            _ctClient.SendGroupingKey("[");
        }

        /*
         * We send the $ and the { keys to Vscode and it automatically
         * adds ${|} to the current cursor location where | represents
         * the cursor..just as if you manually pressed the $ followed
         * by the { key.
         */
        private void grpDollCurlyBraceBtn_Click(object sender, EventArgs e)
        {
            _ctClient.SendGroupingKey("${");
        }

        /*
         * We send the { key to Vscode and it automatically adds
         * {|} to the current cursor location where | represents
         * the cursor..just as if you manually pressed the { key.
         */
        private void grpCurlyBraceBtn_Click(object sender, EventArgs e)
        {
            _ctClient.SendGroupingKey("{");
        }

        /*
         * We send the ( key to Vscode and it automatically adds
         * (|) to the current cursor location where | represents
         * the cursor..just as if you manually pressed the ( key.
         */
        private void grpParenBtn_Click(object sender, EventArgs e)
        {
            _ctClient.SendGroupingKey("(");
        }

        /*
         * Operator Lock buttons - START:
         * 
         * If a operator lock button is already set:
         * 
         * Any multi box left mouse button selections will use the
         * locked operator in building the string to send to the Vscode
         * editor. Even if the Tool Belt is closed. So if the '+'
         * operator box is active and the user lassos box1 and
         * box2 and the box format is currently .to_f() then vscode
         * will receive this string:
         * 
         * "box1.to_f() + box2.to_f()"
         * 
         * If the operator lock button is set [user just clicked it]
         * AND there are boxes in the operator hold textbox we will
         * do the above with those boxes and we will clear the textbox.
         * 
         * If the user clicks a currently locked operator button it will
         * be rest/unlocked and revert back to it's normal color.
         */
        private void boxSelOpLockDivBtn_Click(object sender, EventArgs e)
        {
            ProcessOpLockButton(boxSelOpLock.div);
        }

        private void boxSelOpLockMultBtn_Click(object sender, EventArgs e)
        {
            ProcessOpLockButton(boxSelOpLock.mult);
        }

        private void boxSelOpLockMinusBtn_Click(object sender, EventArgs e)
        {
            ProcessOpLockButton(boxSelOpLock.sub);
        }

        private void boxSelOpLockPlusBtn_Click(object sender, EventArgs e)
        {
            ProcessOpLockButton(boxSelOpLock.add);
        }
        /*
         * Operator Lock buttons - STOP.
         */

        /*
         * Operator/Separator buttons - START:
         * 
         * These perform per click, i.e. not a lock thing.
         * 
         * Refer to ProcessOpSepButton() for details.
         * 
         * Technically possible the count is 1 due to user unselecting
         * boxes that were waiting for an operator/separator selection.
         */
        private void boxSelPlusBtn_Click(object sender, EventArgs e)
        {
            if (_boxSel.toolBeltOpWaitBoxes.Count > 0)
            {
                ProcessOpWaitBoxes((_boxSel.toolBeltOpWaitBoxes.Count > 1 ? " + " : ""));
            }
            else
            {
                _ctClient.SendBoxSelections(" + ");
            }
        }

        private void boxSelMinusBtn_Click(object sender, EventArgs e)
        {
            if (_boxSel.toolBeltOpWaitBoxes.Count > 0)
            {
                ProcessOpWaitBoxes((_boxSel.toolBeltOpWaitBoxes.Count > 1 ? " - " : ""));
            }
            else
            {
                _ctClient.SendBoxSelections(" - ");
            }
        }

        private void boxSelMultBtn_Click(object sender, EventArgs e)
        {
            if (_boxSel.toolBeltOpWaitBoxes.Count > 0)
            {
                ProcessOpWaitBoxes((_boxSel.toolBeltOpWaitBoxes.Count > 1 ? " * " : ""));
            }
            else
            {
                _ctClient.SendBoxSelections(" * ");
            }
        }

        private void boxSelDivBtn_Click(object sender, EventArgs e)
        {
            if (_boxSel.toolBeltOpWaitBoxes.Count > 0)
            {
                ProcessOpWaitBoxes((_boxSel.toolBeltOpWaitBoxes.Count > 1 ? " / " : ""));
            }
            else
            {
                _ctClient.SendBoxSelections(" / ");
            }
        }

        private void boxSelCommaBtn_Click(object sender, EventArgs e)
        {
            if (_boxSel.toolBeltOpWaitBoxes.Count > 0)
            {
                ProcessOpWaitBoxes((_boxSel.toolBeltOpWaitBoxes.Count > 1 ? ", " : ""));
            }
            else
            {
                _ctClient.SendBoxSelections(", ");
            }
        }
        /*
         * Operator/Separator buttons - STOP.
         */

        /*
         * Offering another way to select an operator/separator
         * 'button' for processing the boxes waiting for an
         * operator/separator.
         */
        protected override bool ProcessCmdKey(ref Message msg, Keys keys)
        {
            //MessageBox.Show("Entered ToolBelt's ProcessCmdKey()");

            switch (keys)
            {
                case (Keys.Add):

                    //MessageBox.Show("+");

                    boxSelPlusBtn.PerformClick();

                    break;

                case (Keys.Subtract):

                    //MessageBox.Show("-");

                    boxSelMinusBtn.PerformClick();

                    break;

                case (Keys.Multiply):

                    //MessageBox.Show("*");

                    boxSelMultBtn.PerformClick();

                    break;

                case (Keys.Divide):

                    //MessageBox.Show("/");

                    boxSelDivBtn.PerformClick();

                    break;

                case (Keys.Oemcomma):

                    //MessageBox.Show(",");

                    boxSelCommaBtn.PerformClick();

                    break;

                default:

                    return base.ProcessCmdKey(ref msg, keys);
            }

            return true;
        }

        /*
         * This uses lassoed boxes from the operator wait text box to
         * construct the calc for a variable, basically a line on a
         * form page, say, line1 for example:
         * 
         * line1 = `${box1.to_i()}${box2.to_i()}${box3.to_i()}.${box4.to_i()}`
         * 
         * We can also post from one or more boxes to one or more boxes.
         * In this case the user will have selected 'POST' from the
         * 'Precision' dropdown list. We store the variable name in the
         * variable dictionary along with the number of boxes we are posting
         * to the variable with. We use this value when posting from the 
         * variable for a safety check.
         */
        private void indBoxCalcToVarBtn_Click(object sender, EventArgs e)
        {
            string box = "";
            string indBoxStr = "";
            string msg = "";
            string prec = indBoxPrecComboBox.SelectedItem.ToString();
            string varName = indBoxVarTxtBox.Text;
            int precInt = (prec.IndexOf('.') < 0 ? 0 : prec.Length - (prec.IndexOf('.') + 1));
            int decIndex;
            bool post = (prec == "POST");

            /*
             * Need to verify the number of boxes when precision is > 0
             * which means we have a decimal point and the number of boxes
             * to work with must be at LEAST equal to the precision value.
             */
            if ((precInt > 0) && (_boxSel.toolBeltOpWaitBoxes.Count < precInt))
            {
                msg = string.Format("Individually Boxed Calcs Error:\n\nThe number of boxes must equal or exceed the selected precision.\n\nThe precision is {0} but the number of boxes is {1}.\n\nOperation will be aborted.", precInt, _boxSel.toolBeltOpWaitBoxes.Count);

                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }
            else if (varName.Length == 0)
            {
                msg = "Individually Boxed Calcs Error:\n\nYou must enter a variable name.\n\nOperation will be aborted.";

                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (OrderIndBoxed() == true)
            {
                // Ok to process the boxes - calc to the variable:

                // Debug - START:
                //boxSelTextBox.AppendText("\n\nOrdered Boxes:\n\n");
                //foreach (string box in _boxSel.toolBeltOpWaitBoxes)
                //{
                //    boxSelTextBox.AppendText(string.Format("{0}, ", box));
                //}
                // Debug - STOP:

                if (post == true)
                {
                    /*
                     * We use a different approach for posting. Each box
                     * 'stakes its place' in the string when empty by using
                     * a space char.
                     */
                    indBoxStr = string.Format("{0} = (", varName);

                    for (int i = 0; i < _boxSel.toolBeltOpWaitBoxes.Count; i++)
                    {
                        box = _boxSel.toolBeltOpWaitBoxes[i];

                        if (i == _boxSel.toolBeltOpWaitBoxes.Count - 1)
                        {
                            indBoxStr += string.Format("({0}.isEmpty() ? ' ' : {0})", box);
                        }
                        else
                        {
                            indBoxStr += string.Format("({0}.isEmpty() ? ' ' : {0}) + ", box);
                        }
                    }

                    indBoxStr += string.Format(");{0}{0}", Environment.NewLine);

                    // Update the variable dictionary:
                    indBoxedVarDict[varName] = _boxSel.toolBeltOpWaitBoxes.Count;
                }
                else
                {
                    /*
                     * If there is a decimal point we need to figure it's placement,
                     * amongst other things. We made sure our rule above is met,
                     * that is, the number of boxes MUST be at least as much as the
                     * precision value, so for '9.99' there MUST be at least 2 boxes.
                     * 
                     * Each box 'stakes its place' in the string when empty by using
                     * a 0.
                     */
                    indBoxStr = string.Format("{0} = `", varName);

                    decIndex = (precInt > 0 ? _boxSel.toolBeltOpWaitBoxes.Count - precInt : -1);

                    for (int i = 0; i < _boxSel.toolBeltOpWaitBoxes.Count; i++)
                    {
                        box = _boxSel.toolBeltOpWaitBoxes[i];

                        if (i == decIndex)
                        {
                            indBoxStr += string.Format(".${{{0}.to_i()}}", box);
                        }
                        else
                        {
                            indBoxStr += string.Format("${{{0}.to_i()}}", box);
                        }
                    }

                    indBoxStr += string.Format("`;{0}{0}", Environment.NewLine);
                }

                // Debug:
                //boxSelTextBox.AppendText(string.Format("\n\n{0}", indBoxStr));

                /*
                 * Reset the Box Selector's operator wait state. We will leave
                 * the variable name text box as is to reduce consecutive
                 * individually boxed calculations having to retype a common
                 * variable name, for example, 'line1', then 'line2', then 'line3'
                 * and so on...just change the '1' to a '2' to a '3'.
                 */
                ResetBoxSelOpWait(true);

                _ctClient.SendBoxSelections(indBoxStr);
            }
        }

        /*
         * This uses lassoed boxes from the operator wait text box to
         * construct the calc from a variable - so for example, we would
         * generate the code that would distribute the value of a variable,
         * say, 'subtotal', to the lassoed boxes.
         */
        private void indBoxCalcFromVarBtn_Click(object sender, EventArgs e)
        {
            string msg = "";
            string prec = indBoxPrecComboBox.SelectedItem.ToString();
            string varName = indBoxVarTxtBox.Text;
            bool post = (prec == "POST");
            int precInt = (prec.IndexOf('.') < 0 ? 0 : prec.Length - (prec.IndexOf('.') + 1));
            int decIndex;
            int boxIndex = 0;
            int formatLen;
            int ifCnt = 0;
            string indBoxStr = "";
            string ifStr = "";
            string valueStr = "";
            string assignStr = "";
            string printStr = "";
            string box = "";

            /*
             * Need to verify the number of boxes when precision is > 0
             * which means we have a decimal point and the number of boxes
             * to work with must be at LEAST equal to the precision value.
             */
            if ((precInt > 0) && (_boxSel.toolBeltOpWaitBoxes.Count < precInt))
            {
                msg = string.Format("Individually Boxed Calcs Error:\n\nThe number of boxes must equal or exceed the selected precision.\n\nThe precision is {0} but the number of boxes is {1}.\n\nOperation will be aborted.", precInt, _boxSel.toolBeltOpWaitBoxes.Count);

                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }
            else if (varName.Length == 0)
            {
                msg = "Individually Boxed Calcs Error:\n\nYou must enter a variable name.\n\nOperation will be aborted.";

                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (OrderIndBoxed() == true)
            {
                // Ok to process the boxes - calc from the variable:
                formatLen = (precInt > 0 ? _boxSel.toolBeltOpWaitBoxes.Count + 1 : _boxSel.toolBeltOpWaitBoxes.Count);

                if (post == true)
                {
                    /* 
                     * Checking if the variable name is in the dictionary, if so we compare
                     * the variable's box length [the number of boxes used in calculating to
                     * the variable] to the number of boxes used in calculating from the variable.
                     * We want to warn the user since this is a post, basically copying from one
                     * area on a document to another area, we want to get a complete copy.
                     */
                    if (indBoxedVarDict.ContainsKey(varName) && (indBoxedVarDict[varName] != formatLen))
                    {
                        int varCnt = indBoxedVarDict[varName];

                        msg = string.Format("Individually Boxed Calcs Warning:\n\n{0} box{1} {2} posted to {3} however {4} box{5} {6} being posted from {3}.\n\nMake sure you have properly selected your boxes for this posting operation.", varCnt, (varCnt > 1 ? "es" : ""), (varCnt > 1 ? "were" : "was"), varName, formatLen, (formatLen > 1 ? "es" : ""), (formatLen > 1 ? "are" : "is"));

                        using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.warning, msg))
                        {
                            mainMsg.ShowDialog(this);
                        }
                    }

                    valueStr = string.Format("value = {0}.padEnd({1});{2}{2}", varName, formatLen, Environment.NewLine);
                }
                else
                {
                    /*
                     * Here we will look for a box as the 'Calc From Variable' and if true
                     * limit it's value to the maximum value that can be represented by the
                     * number of destination boxes.
                     */
                    Regex boxVar = new Regex((_docFile.newBoxFormat == true ? @"^box\d+$" : @"^box\d+_\d+$"));
                    string maxValue = "";

                    if (boxVar.IsMatch(varName))
                    {
                        maxValue = (new String('9', _boxSel.toolBeltOpWaitBoxes.Count - precInt)) + (precInt > 0 ? "." : "") + (new String('9', precInt));
                        valueStr = string.Format("value = Math.min({0}.to_f(), {1}).complexFixedString({2}, '{3}');{4}{4}", varName, maxValue, precInt, formatLen, Environment.NewLine);
                    }
                    else
                    {
                        valueStr = string.Format("value = {0}.complexFixedString({1}, '{2}');{3}{3}", varName, precInt, formatLen, Environment.NewLine);
                    }
                }

                /*
                 * If there is a decimal point we need to figure it's placement,
                 * amongst other things. We made sure our rule above is met,
                 * that is, the number of boxes MUST be at least as much as the
                 * precision value, so for '9.99' there MUST be at least 2 boxes.
                 */
                decIndex = (precInt > 0 ? _boxSel.toolBeltOpWaitBoxes.Count - precInt : -1);
                ifStr = "if (";

                for (int i = 0; i < formatLen; i++)
                {
                    if (i != decIndex)
                    {
                        box = _boxSel.toolBeltOpWaitBoxes[boxIndex];
                        boxIndex += 1;
                    }

                    // The if (hasOverrides) string:
                    if (i != decIndex)
                    {
                        if (ifCnt > 0)
                        {
                            ifStr += string.Format(" &&{0}(!SFForms.StorageManager.hasOverrides(formInstanceId, \"{1}\"))", Environment.NewLine, box);
                        }
                        else
                        {
                            ifCnt += 1;
                            ifStr += string.Format("(!SFForms.StorageManager.hasOverrides(formInstanceId, \"{0}\"))", box);
                        }
                    }

                    // The assignStr string:
                    if (i == decIndex)
                    {
                        assignStr += string.Format("// Jump over the decimal point.{0}", Environment.NewLine);
                    }
                    else
                    {
                        assignStr += string.Format("{0} = value.substr(value.length - {1}, 1);{2}", box, formatLen - i, Environment.NewLine);
                    }

                    // The printStr string:
                    if (i != decIndex)
                    {
                        printStr += string.Format("SFForms.StorageManager.setValue(formInstanceId, '{0}', {0});{1}", box, Environment.NewLine);
                    }
                }

                // Close the if (hasOverrides) string:
                ifStr += string.Format(") {{{0}{0}", Environment.NewLine);

                // Now put it all together:
                indBoxStr = string.Format("{0}{1}{2}}}{3}{3}{4}{3}{3}", ifStr, valueStr, assignStr, Environment.NewLine, printStr);
                //indBoxStr = string.Format("{0}{1}{2}}}{3}{4}{5}{6}{7}", ifStr, valueStr, assignStr, Environment.NewLine, Environment.NewLine, printStr, Environment.NewLine, Environment.NewLine);

                // Debug:
                //boxSelTextBox.AppendText(string.Format("\n\n{0}", indBoxStr));

                /*
                 * Reset the Box Selector's operator wait state. We will leave
                 * the variable name text box as is to reduce consecutive
                 * individually boxed calculations having to retype a common
                 * variable name, for example, 'line1', then 'line2', then 'line3'
                 * and so on...just change the '1' to a '2' to a '3'.
                 */
                ResetBoxSelOpWait(true);

                _ctClient.SendBoxSelections(indBoxStr);
            }
        }

        /*
         * This is called when the user chooses to perform an individually
         * boxed calculation. The box list _boxSel.toolBeltOpWaitBoxes is
         * processed for this task.
         * 
         * As an example, here is a 'Calc To Variable' calculation:
         * 
         * line1 = `${box1.to_i()}${box2.to_i()}${box3.to_i()}.${box4.to_i()}`
         * 
         * We 'order' the boxes from left to right using the field left
         * coordinates, and also verify all boxes are on the same 'row' on
         * the line, i.e. no vertically positioned boxes [relative to one
         * another] will be allowed, and the above example must be what
         * would be the case on the pdf page, box1 followed by box2,
         * and so on.
         * 
         * Not going nuts here, but good enough for a reasonable person
         * to be able to handle box selections for an individually
         * boxed calculation, i.e. you don't have to be an Einstein here.
         * 
         * Called From:
         * 
         *      indBoxCalcToVarBtn_Click()
         */
        public bool OrderIndBoxed()
        {
            bool result = true;
            List<float> tmpList = new List<float>();
            List<string> tmpBoxList = new List<string>();
            Dictionary<float, IndBoxed> indBoxedDict = new Dictionary<float, IndBoxed>();
            IndBoxed indBoxed;
            List<Field> fa = null;
            LeftBoxSel lbSel;
            string errMsg = "";
            int curPaIndex = -1;
            float firstTop;
            float firstBottom;
            float avgTopBot;

            foreach (string box in _boxSel.toolBeltOpWaitBoxes)
            {
                /*
                 * Idiot check #1 - all boxes must be in the leftBoxSelDict!
                 */
                if (_boxSel.leftBoxSelDict.TryGetValue(box, out lbSel) == false)
                {
                    errMsg = string.Format("Individually Boxed Calcs Error:\n\n{0} is not in the leftBoxSelDict.\n\nAborting operation.", box);
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                    {
                        mainMsg.ShowDialog(this);
                    }

                    result = false;
                    break;
                }

                /*
                 * Idiot check #2 - all boxes must be on same page!
                 */
                if (lbSel.paIndex != curPaIndex)
                {
                    if (curPaIndex != -1)
                    {
                        errMsg = "Individually Boxed Calcs Error:\n\nAn individually boxed calculation can only be performed when boxes are on the same page.\n\nAborting operation.";

                        using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                        {
                            mainMsg.ShowDialog(this);
                        }

                        result = false;
                        break;
                    }
                    else
                    {
                        _pa.GetFieldArrayRef(lbSel.paIndex, ref fa);
                        curPaIndex = lbSel.paIndex;

                        /*
                         * Idiot check #3 - got a null field array!
                         */
                        if (fa == null)
                        {
                            errMsg = string.Format("Individually Boxed Calcs Error:\n\n_pa.GetFieldArrayRef() returned null field array while processing {0}.\n\nAborting operation.", box);

                            using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                            {
                                mainMsg.ShowDialog(this);
                            }

                            result = false;
                            break;
                        }
                    }
                }

                /*
                 * Add to our local dictionary. We will sort based on the
                 * key, which is the left field coordinate. Duplicate left
                 * field coordinates means there is an error, either a box
                 * colocated on top of another [same box?] where the left
                 * sides are the same, or more likely boxes are on different
                 * rows of the form but are lined up on the left side - a 
                 * show stopper for us.
                 * 
                 * If we pass this, then we can simply sort the list and get the
                 * proper left to right order of boxes for our calc.
                 * 
                 * Idiot check #4 - no left side duplicates!
                 */
                if (indBoxedDict.ContainsKey(fa[lbSel.faIndex].bbox.left))
                {
                    errMsg = "Individually Boxed Calcs Error:\n\nOne or more boxes share the same left side coordinate.\n\nAborting Operation.";
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                    {
                        mainMsg.ShowDialog(this);
                    }

                    result = false;
                    break;
                }

                indBoxed = new IndBoxed(box, fa[lbSel.faIndex].bbox.top, fa[lbSel.faIndex].bbox.bottom);
                indBoxedDict.Add(fa[lbSel.faIndex].bbox.left, indBoxed);
            }

            if (result == true)
            {
                tmpList = indBoxedDict.Keys.ToList();
                tmpList.Sort();

                /*
                 * Ok, so we just did the left to right box verification, now we
                 * do the up/down/vertical box verification:
                 * 
                 * Rebuild the box list, and make sure the boxes are on the same 'row',
                 * i.e. cannot have one box on line 1 and another on line 2, etc..
                 */
                tmpBoxList.Add(indBoxedDict[tmpList[0]].box);
                firstTop = indBoxedDict[tmpList[0]].top;
                firstBottom = indBoxedDict[tmpList[0]].bottom;

                for (int i = 1; i < tmpList.Count; i++)
                {
                    avgTopBot = (indBoxedDict[tmpList[i]].top + indBoxedDict[tmpList[i]].bottom) / 2.0f;

                    if ((avgTopBot < firstTop) && (avgTopBot > firstBottom))
                    {
                        tmpBoxList.Add(indBoxedDict[tmpList[i]].box);
                    }
                    else
                    {
                        errMsg = string.Format("Individually Boxed Calcs Error:\n\nBoxes are not aligned vertically. \n\n{0} vertical position differs substantially from {1}.\n\nAborting operation.", tmpBoxList[0], indBoxedDict[tmpList[i]].box);

                        using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                        {
                            mainMsg.ShowDialog(this);
                        }

                        result = false;
                        break;
                    }
                }

                if (result == true)
                {
                    /*
                     * We made it! Now clear and re-populate _boxSel.toolBeltOpWaitBoxes.
                     */
                    _boxSel.toolBeltOpWaitBoxes.Clear();
                    _boxSel.toolBeltOpWaitBoxes.AddRange(tmpBoxList);
                }
            }

            return result;
        }

        /*
         * *****************
         * Snippet Picker:
         * 
         *  List selections.
         */
        private void snipCatList_SelectedIndexChanged(object sender, EventArgs e)
        {
            string key;

            if (snipCatList.SelectedIndex != -1)
            {
                snipPrefixList.Items.Clear();

                key = _docFile.snipPickDict.Keys.ElementAt(snipCatList.SelectedIndex);

                foreach (string prefix in _docFile.snipPickDict[key].snipPrefixes)
                {
                    snipPrefixList.Items.Add(prefix);
                }

                if (snipPrefixList.Items.Count > 0)
                {
                    snipPrefixList.SetSelected(0, true);
                }
            }
        }

        private void snipPrefixList_SelectedIndexChanged(object sender, EventArgs e)
        {
            string key;

            if (snipPrefixList.SelectedIndex != -1)
            {
                snipDescTextBox.Clear();
                snipCodeTextBox.Clear();

                key = _docFile.snipPickDict.Keys.ElementAt(snipCatList.SelectedIndex);

                snipDescTextBox.Text = _docFile.snipPickDict[key].snipDescs[snipPrefixList.SelectedIndex].Replace("\\n", Environment.NewLine);

                snipCodeTextBox.Text = _docFile.snipPickDict[key].snipBodys[snipPrefixList.SelectedIndex].Replace("\\n", Environment.NewLine);
            }
        }

        /*
         *  Button clicks.
         */
        private void snipInsertBtn_Click(object sender, EventArgs e)
        {
            string key;
            string errMsg;

            if (_ctClient.boxSelClient == boxSelDest.document)
            {
                if ((snipCatList.SelectedIndex != -1) && (snipPrefixList.SelectedIndex != -1))
                {
                    key = _docFile.snipPickDict.Keys.ElementAt(snipCatList.SelectedIndex);

                    _ctClient.InsertSnippet(_docFile.snipPickDict[key].snipNames[snipPrefixList.SelectedIndex]);
                }
                else
                {
                    errMsg = "Insert Snippet Error:\n\nYou must have an active selection in both the 'Snippet Categories' and the 'Snippet Prefix' lists before you can insert a snippet.";

                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                    {
                        mainMsg.ShowDialog(this);
                    }
                }
            }
            else
            {
                errMsg = "Insert Snippet Error:\n\nCalc Helper must be in Document Box Selection mode.\n\nAborting Operation.";

                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                {
                    mainMsg.ShowDialog(this);
                }
            }
        }

        /*
         * *****************
         * Function Picker:
         * 
         *  List selection.
         */
        private void fcnNameList_SelectedIndexChanged(object sender, EventArgs e)
        {
            string key;

            if (fcnNameList.SelectedIndex != -1)
            {
                fcnDescTextBox.Clear();
                fcnCodeTextBox.Clear();
                fcnCallDescTextBox.Clear();
                fcnCallTextBox.Clear();

                key = _docFile.fcnSnipPickDict.Keys.ElementAt(fcnNameList.SelectedIndex);

                fcnDescTextBox.Text = _docFile.fcnSnipPickDict[key].fcnDesc.Replace("\\n", Environment.NewLine);
                fcnCodeTextBox.Text = _docFile.fcnSnipPickDict[key].fcnBody.Replace("\\n", Environment.NewLine);
                fcnCallDescTextBox.Text = _docFile.fcnSnipPickDict[key].fcnCallDesc.Replace("\\n", Environment.NewLine);
                fcnCallTextBox.Text = _docFile.fcnSnipPickDict[key].fcnCallBody.Replace("\\n", Environment.NewLine);
            }
        }

        /*
         *  Button clicks.
         */
        private void insertFcnCodeBtn_Click(object sender, EventArgs e)
        {
            string key;
            string errMsg;

            if (_ctClient.boxSelClient == boxSelDest.document)
            {
                if (fcnNameList.SelectedIndex != -1)
                {
                    key = _docFile.fcnSnipPickDict.Keys.ElementAt(fcnNameList.SelectedIndex);

                    _ctClient.InsertSnippet(_docFile.fcnSnipPickDict[key].fcnName);
                }
                else
                {
                    errMsg = "Insert Function Snippet Error:\n\nYou must have an active selection in the 'Function' list before you can insert a function snippet.";

                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                    {
                        mainMsg.ShowDialog(this);
                    }
                }
            }
            else
            {
                errMsg = "Insert Function Snippet Error:\n\nCalc Helper must be in Document Box Selection mode.\n\nAborting Operation.";

                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                {
                    mainMsg.ShowDialog(this);
                }
            }
        }

        private void insertFcnCallBtn_Click(object sender, EventArgs e)
        {
            string key;
            string errMsg;

            if (_ctClient.boxSelClient == boxSelDest.document)
            {
                if (fcnNameList.SelectedIndex != -1)
                {
                    key = _docFile.fcnSnipPickDict.Keys.ElementAt(fcnNameList.SelectedIndex);

                    _ctClient.InsertSnippet(_docFile.fcnSnipPickDict[key].fcnCallName);
                }
                else
                {
                    errMsg = "Insert Function Call Snippet Error:\n\nYou must have an active selection in the 'Function' list before you can insert a function call snippet.";

                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                    {
                        mainMsg.ShowDialog(this);
                    }
                }
            }
            else
            {
                errMsg = "Insert Function Call Snippet Error:\n\nCalc Helper must be in Document Box Selection mode.\n\nAborting Operation.";

                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                {
                    mainMsg.ShowDialog(this);
                }
            }
        }

        /*
         * ***************
         * Setting focus to the ToolBelt: 
         * 
         * When someone directly wants to Exit the ToolBelt for example.
         */
        private void toolStrip1_MouseEnter(object sender, EventArgs e)
        {
            this.Focus();
        }

        // When user mouses into main ToolBelt form body:
        private void ToolBelt_MouseEnter(object sender, EventArgs e)
        {
            this.Focus();
        }

        // When user mouses into main ToolBelt form body:
        private void toolBeltTabCtrl_MouseEnter(object sender, EventArgs e)
        {
            this.Focus();
        }
        /*
         * ***************
         */
    }

    // Things we tell the form about:
    public enum tbEvent { clearAllSel, clearOpWaitSel, clearBoxSel };

    public class ToolBeltArgs : EventArgs
    {
        public tbEvent eventType { get; set; }
        public string eventMsg { get; set; }
        public List<string> boxSel { get; set; }
    }

    // For individually boxed calcs.
    public class IndBoxed
    {
        public string box;
        public float top;
        public float bottom;

        public IndBoxed(string indBox, float indTop, float indBottom)
        {
            box = indBox;
            top = indTop;
            bottom = indBottom;
        }
    }

}
